package org.netkernel.client.http.endpoint;


import javax.net.ssl.SSLContext;

import org.apache.http.config.Registry;
import org.apache.http.config.RegistryBuilder;
import org.apache.http.conn.socket.ConnectionSocketFactory;
import org.apache.http.conn.socket.PlainConnectionSocketFactory;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.apache.http.impl.nio.conn.PoolingNHttpClientConnectionManager;
import org.apache.http.impl.nio.reactor.DefaultConnectingIOReactor;
import org.apache.http.impl.nio.reactor.IOReactorConfig;
import org.apache.http.nio.reactor.ConnectingIOReactor;
import org.netkernel.client.http.representation.IHttpClientConfig;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;


public class DefaultConnectionManagerAccessor extends StandardAccessorImpl
{
	public static final int MAXIMUM_TOTAL_HTTP_CLIENT_CONNECTIONS=64;
	public static final int MAXIMUM_TOTAL_HTTP_CLIENT_CONNECTIONS_PER_HOST=8;
	public static final String MAXIMUM_TOTAL_HTTP_CLIENT_CONNECTIONS_PER_HOST_PROPERTY="http.client.max.total.connections.per.host";
	public static final String MAXIMUM_TOTAL_HTTP_CLIENT_CONNECTIONS_PROPERTY="http.client.max.total.connections";
	public static final String HTTP_ASYNC_CLIENT_REACTOR_THREADS="http.client.async.reactor.threads";
	
	public DefaultConnectionManagerAccessor()
	{	this.declareThreadSafe();		
	}

	@Override
	public void onSource(INKFRequestContext aContext) throws Exception
	{	int connsPerHost, totalConns;
		try
		{ totalConns=aContext.getKernelContext().getKernel().getConfiguration().getInt(MAXIMUM_TOTAL_HTTP_CLIENT_CONNECTIONS_PROPERTY);
		}
		catch(Exception e)
		{	totalConns=MAXIMUM_TOTAL_HTTP_CLIENT_CONNECTIONS;		
		}
		try
		{ connsPerHost=aContext.getKernelContext().getKernel().getConfiguration().getInt(MAXIMUM_TOTAL_HTTP_CLIENT_CONNECTIONS_PER_HOST_PROPERTY);
		}
		catch(Exception e)
		{	connsPerHost=MAXIMUM_TOTAL_HTTP_CLIENT_CONNECTIONS_PER_HOST;			
		}
		
		Object cm=null;
		if(aContext.getThisRequest().getIdentifier().startsWith("active:httpAsync"))
		{	
			IHttpClientConfig key=aContext.source("arg:key", IHttpClientConfig.class);
			{
				PoolingNHttpClientConnectionManager m=null;
				// Create I/O reactor configuration
				int reactorThreads;
				try
				{ reactorThreads=aContext.getKernelContext().getKernel().getConfiguration().getInt(HTTP_ASYNC_CLIENT_REACTOR_THREADS);
				}
				catch(Exception e)
				{	reactorThreads=Runtime.getRuntime().availableProcessors();		
				}
		        IOReactorConfig ioReactorConfig = IOReactorConfig.custom()
		                .setIoThreadCount(reactorThreads)
		                .build();
		        // Create a custom I/O reactor
		        ConnectingIOReactor ioReactor = new DefaultConnectingIOReactor(ioReactorConfig);
	
		        // Create a connection manager with custom configuration.
		        m = new PoolingNHttpClientConnectionManager(ioReactor);
		        m.setDefaultMaxPerRoute(connsPerHost);
		        m.setMaxTotal(totalConns);
		        cm=m;
			}
		}
		else
		{
			final SSLContext sslcontext=SSLContext.getInstance("TLSv1.2");
			sslcontext.init(null, null, null);
			Registry<ConnectionSocketFactory> r = RegistryBuilder.<ConnectionSocketFactory>create()
			        .register("http", PlainConnectionSocketFactory.getSocketFactory())
			        .register("https", new SSLConnectionSocketFactory(sslcontext))
			        .build();
			PoolingHttpClientConnectionManager m = new PoolingHttpClientConnectionManager(r);
			
			m.setMaxTotal(totalConns);
			m.setDefaultMaxPerRoute(connsPerHost);
			cm=m;
		}
		
		aContext.createResponseFrom(cm);
	}
}
