package org.netkernel.client.http.endpoint;

import oauth.signpost.OAuthConsumer;
import oauth.signpost.commonshttp.CommonsHttpOAuthConsumer;

import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.HttpRequest;
import org.apache.http.HttpResponse;
import org.apache.http.client.config.CookieSpecs;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.protocol.HttpClientContext;
import org.apache.http.impl.cookie.DateParseException;
import org.apache.http.impl.cookie.DateUtils;
import org.apache.http.message.BasicHeader;
import org.netkernel.client.http.representation.IHttpClientConfig;
import org.netkernel.client.http.representation.OAuthCredentials;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.representation.ByteArrayRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Date;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


public class HttpClientAccessorBase extends StandardAccessorImpl
{	
	private static HttpClientAccessorBase sInstance;
	
	private static final String NETKERNEL_USER_AGENT="1060 NetKernel HTTP Client 2.0.0  http://www.1060research.com";
	
	protected static final String HTTP_ACCESSOR_STATUS_CODE_METADATA="HTTP_ACCESSOR_STATUS_CODE_METADATA";
	protected static final String HTTP_ACCESSOR_RESPONSE_HEADERS_METADATA="HTTP_ACCESSOR_RESPONSE_HEADERS_METADATA";
	
	protected static final String HTTP_METHOD_GET="httpGet";
	protected static final String HTTP_METHOD_GET_ASYNC="httpAsyncGet";
	protected static final int HTTP_METHOD_GET_INT=1;
	protected static final String HTTP_METHOD_PUT="httpPut";
	protected static final String HTTP_METHOD_PUT_ASYNC="httpAsyncPut";
	protected static final int HTTP_METHOD_PUT_INT=2;
	protected static final String HTTP_METHOD_POST="httpPost";
	protected static final String HTTP_METHOD_POST_ASYNC="httpAsyncPost";
	protected static final int HTTP_METHOD_POST_INT=3;
	protected static final String HTTP_METHOD_HEAD="httpHead";
	protected static final String HTTP_METHOD_HEAD_ASYNC="httpAsyncHead";
	protected static final int HTTP_METHOD_HEAD_INT=4;
	protected static final String HTTP_METHOD_DELETE="httpDelete";
	protected static final String HTTP_METHOD_DELETE_ASYNC="httpAsyncDelete";
	protected static final int HTTP_METHOD_DELETE_INT=5;
	protected static final String HTTP_METHOD_PATCH="httpPatch";
	protected static final String HTTP_METHOD_PATCH_ASYNC="httpAsyncPatch";
	protected static final int HTTP_METHOD_PATCH_INT=6;
	protected static final String HTTP_METHOD_OPTIONS="httpOptions";
	protected static final String HTTP_METHOD_OPTIONS_ASYNC="httpAsyncOptions";
	protected static final int HTTP_METHOD_OPTIONS_INT=7;
	
	
	protected static final String CONTENT_TYPE_HEADER="Content-Type";
	protected static final String EXPIRES_HEADER="Expires";
	protected static final String CACHE_CONTROL_HEADER="Cache-Control";
	
	
	private static Pattern mMaxAgePattern=Pattern.compile("max-age.*?=.*?([0-9]+)",Pattern.DOTALL|Pattern.MULTILINE);
	
	//private static final String HTTP_CONFIG_URI="res:/etc/HTTPClientConfig.xml";
	protected static final String HTTP_CONFIG_URI_DEFAULT="res:/etc/DefaultHTTPClientConfig.xml";
	
	protected static final HashMap mModes=new HashMap();
	
	//public static final SchemeRegistry mHTTPClientSchemeRegistry = new SchemeRegistry();
	
	{
		mModes.put(HTTP_METHOD_GET, new Integer(HTTP_METHOD_GET_INT));
		mModes.put(HTTP_METHOD_GET_ASYNC, new Integer(HTTP_METHOD_GET_INT));
		mModes.put(HTTP_METHOD_PUT, new Integer(HTTP_METHOD_PUT_INT));		
		mModes.put(HTTP_METHOD_PUT_ASYNC, new Integer(HTTP_METHOD_PUT_INT));
		mModes.put(HTTP_METHOD_POST, new Integer(HTTP_METHOD_POST_INT));
		mModes.put(HTTP_METHOD_POST_ASYNC, new Integer(HTTP_METHOD_POST_INT));
		mModes.put(HTTP_METHOD_HEAD, new Integer(HTTP_METHOD_HEAD_INT));
		mModes.put(HTTP_METHOD_HEAD_ASYNC, new Integer(HTTP_METHOD_HEAD_INT));
		mModes.put(HTTP_METHOD_OPTIONS, new Integer(HTTP_METHOD_OPTIONS_INT));
		mModes.put(HTTP_METHOD_OPTIONS_ASYNC, new Integer(HTTP_METHOD_OPTIONS_INT));
		mModes.put(HTTP_METHOD_DELETE, new Integer(HTTP_METHOD_DELETE_INT));
		mModes.put(HTTP_METHOD_DELETE_ASYNC, new Integer(HTTP_METHOD_DELETE_INT));
		mModes.put(HTTP_METHOD_PATCH, new Integer(HTTP_METHOD_PATCH_INT));
		mModes.put(HTTP_METHOD_PATCH_ASYNC, new Integer(HTTP_METHOD_PATCH_INT));
	}
	
	/** Creates a new instance of HTTPClientAccessor */
	public HttpClientAccessorBase()
	{	declareThreadSafe();
		System.getProperties().setProperty("httpclient.useragent", NETKERNEL_USER_AGENT);
		Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());

		sInstance=this;
	}
	
	public static HttpClientAccessorBase getInstance()
	{	return sInstance;
	}
	

	@Override
	public void onSource(INKFRequestContext aContext) throws Exception
	{	throw new Exception("Must be overridden");	

	}
	
	protected String escapeURL(String aURL)
	{	String result=aURL;
		boolean illegal=false;
		for (int i=0; i<aURL.length() && !illegal; i++)
		{	char c=aURL.charAt(i);
			if (isIllegalURIChar(c)) illegal=true;
		}
		if (illegal)
		{	StringBuilder sb=new StringBuilder(aURL.length()*2);
			for (int i=0; i<aURL.length(); i++)
			{	char c=aURL.charAt(i);
				if (isIllegalURIChar(c))
				{	sb.append('%');
					sb.append(Integer.toHexString(c));
				}
				else
				{	sb.append(c);
				}
			}
			result=sb.toString();
		}
		return result;
	}
	
	//according to http://www.ietf.org/rfc/rfc2396.txt
	protected static String SAFE="_-!.~'()*,;:$&+=?/[]@%";
	protected static boolean isIllegalURIChar(char c)
	{	boolean other=c>127 && !Character.isISOControl(c) && !Character.isSpaceChar(c);
		boolean safe=Character.isLetterOrDigit(c) || SAFE.indexOf(c)>=0 || other;
		return !safe;
	}
	
	protected void oauthSignRequest(HttpRequest request, OAuthCredentials credentials) throws Exception
	{	OAuthConsumer consumer=new CommonsHttpOAuthConsumer(credentials.getConsumerKey(), credentials.getConsumerSecret());
		consumer.setTokenWithSecret(credentials.getAccessToken(), credentials.getAccessTokenSecret());
		consumer.sign(request);
	}
	
	
	protected IHDSNode getResponseHeaders(Header[] aHeaders)
	{	HDSBuilder b=new HDSBuilder();
		for(Header h: aHeaders)
		{	b.addNode(h.getName(),h.getValue());
		}
		return b.getRoot();
	}
		
	protected void setRequestHeaders(HttpRequest method, IHDSNode headers)
	{	
		//Be tolerant and do a little search to find a suitable header list.  Therefore headers argument can be flat HDS or root noded HDS or NVP tree.
		IHDSNode[] nl=headers.getChildren();
		if(nl.length==1)
		{	String root=nl[0].getName();
			if(root!=null)
			{	String upperRoot=root.toUpperCase();
				if(upperRoot.equals("HDS") || upperRoot.equals("NVP"))
				{	headers=nl[0];
				}
			}
		}
		for(IHDSNode n: headers.getChildren())
		{	String value=(String)n.getValue();
			method.setHeader(n.getName(),value);
		}
	}
	
	protected ByteArrayRepresentation getRepresentationFromHttpResponse(HttpResponse response) throws Exception
	{	HttpEntity entity=response.getEntity();
		byte[] result;
		if(entity!=null)
		{	Long contentLength=entity.getContentLength();
			if(contentLength<0)
			{ contentLength=2048L;
			}
			ByteArrayOutputStream baos=new ByteArrayOutputStream(contentLength.intValue());
			response.getEntity().writeTo(baos);
			result=baos.toByteArray();
		}
		else
		{	//Happens with 304 response
			result=new byte[0];			
		}
		return new ByteArrayRepresentation(result);
	}
	
	protected String getContentType(Header h)
	{	String result;
		if(h!=null)
		{	String val=h.getValue();
			int i=val.indexOf(';');
			if(i>0)	result=val.substring(0,i);
			else result=val;
		}
		else result="content/unknown";
		return result;
	}
	
	protected long getExpires(HttpResponse aResponse)
	{	long expires=0;
		try
		{	Header h=aResponse.getFirstHeader(CACHE_CONTROL_HEADER);
			if(h!=null)
			{	String v=h.getValue();
				Matcher m=mMaxAgePattern.matcher(v);
				if(m.find())
				{	String ma=m.group(1);
					expires=Integer.parseInt(ma)*1000L+System.currentTimeMillis();
					//Take this even it is greater than any absolute Expires time due to this rule
					/*
					 * http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.9.3
					 * 
					 * If a response includes both an Expires header and a max-age directive,
					 * the max-age directive overrides the Expires header, even if the Expires
					 * header is more restrictive. This rule allows an origin server to provide,
					 * for a given response, a longer expiration time to an HTTP/1.1 (or later)
					 * cache than to an HTTP/1.0 cache. This might be useful if certain HTTP/1.0
					 * caches improperly calculate ages or expiration times, perhaps
					 * due to desynchronized clocks. 
					 */
					
				}
			}
		}
		catch(NumberFormatException e){/*Must be expired*/}
		if(expires==0)
		{	//Look at the Expires header since there's no Cache-Control
			try
			{	Header h=aResponse.getFirstHeader(EXPIRES_HEADER);
				if(h!=null)
				{	Date d2=DateUtils.parseDate( h.getValue());
					expires=d2.getTime();
				}
			}
			catch(DateParseException e){/*Must be expired*/}
		}
		return expires;
	}
	
	protected RequestConfig.Builder getAndIntitiallyConfigureRequestConfigBuilder(IHttpClientConfig cfg, HttpClientContext state, boolean entityEnclosingMethod, INKFRequestContext aContext, String aURL)
	{
		RequestConfig.Builder rqb=RequestConfig.custom();
		if(!entityEnclosingMethod)
		{	rqb.setRedirectsEnabled(cfg.followRedirects());
		}
		else
		{	//Entity enclosing methods cannot follow redirects without user intervention
			rqb.setRedirectsEnabled(false);
		}
		rqb.setConnectionRequestTimeout(cfg.getConnectTimeout());
		rqb.setConnectTimeout(cfg.getConnectTimeout());
		rqb.setSocketTimeout(cfg.getTimeout());
		
		if(!aContext.getThisRequest().argumentExists("state"))
		{	//Turn off cookie processing as we have no persistent state
			rqb.setCookieSpec(CookieSpecs.IGNORE_COOKIES);
		}
		else
		{	rqb.setCookieSpec(CookieSpecs.BEST_MATCH);
		}
		//Deal with Client specific settings
		rqb.setExpectContinueEnabled(cfg.expectContinue());
		return rqb;
	}
	
	protected class ReadableBinaryRequestEntity implements HttpEntity
	{	
		private IReadableBinaryStreamRepresentation mRBS;
		
		public void consumeContent() throws IOException
		{
		}

		public InputStream getContent() throws IOException, IllegalStateException
		{	return mRBS.getInputStream();
		}

		public Header getContentEncoding()
		{	return new BasicHeader("Content-Encoding", mRBS.getEncoding());
		}

		public boolean isChunked()
		{	return false;
		}

		public boolean isStreaming()
		{	return true;
		}

		public void writeTo(OutputStream out) throws IOException
		{	mRBS.write(out);
		}

		private String mType;
		
		public ReadableBinaryRequestEntity(IReadableBinaryStreamRepresentation aRBS, String aType)
		{	mRBS=aRBS;
			mType=aType;
		}

		public long getContentLength()
		{	return (long)mRBS.getContentLength();
		}

		public Header getContentType()
		{	return new BasicHeader(HttpClientAccessorBase.CONTENT_TYPE_HEADER, mType);
		}

		public boolean isRepeatable()
		{	return true;
		}
	}
}
