package org.netkernel.client.http.endpoint;


import org.apache.commons.httpclient.methods.HttpPatch;
import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.HttpHost;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.NameValuePair;
import org.apache.http.auth.AuthOption;
import org.apache.http.auth.AuthScheme;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.Credentials;
import org.apache.http.auth.MalformedChallengeException;
import org.apache.http.auth.NTCredentials;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.AuthenticationStrategy;
import org.apache.http.client.CookieStore;
import org.apache.http.client.CredentialsProvider;
import org.apache.http.client.config.AuthSchemes;
import org.apache.http.client.config.CookieSpecs;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpHead;
import org.apache.http.client.methods.HttpOptions;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.protocol.HttpClientContext;
import org.apache.http.conn.HttpClientConnectionManager;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.impl.client.BasicCredentialsProvider;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.DefaultHttpRequestRetryHandler;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.nio.client.CloseableHttpAsyncClient;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.protocol.HttpContext;
import org.apache.http.util.EntityUtils;
import org.netkernel.client.http.representation.HttpCacheRepresentation;
import org.netkernel.client.http.representation.HttpClientResponseRepresentation;
import org.netkernel.client.http.representation.HttpStateRepresentation;
import org.netkernel.client.http.representation.IHttpClientConfig;
import org.netkernel.client.http.representation.OAuthCredentials;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.representation.ByteArrayRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;

import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.regex.Pattern;


public class HttpClientAccessorImpl extends HttpClientAccessorBase
{	
	private static HttpClientAccessorImpl sInstance;
	
	/** Creates a new instance of HTTPClientAccessor */
	public HttpClientAccessorImpl()
	{	super();
		sInstance=this;
	}
	
	public static HttpClientAccessorImpl getInstance()
	{	return sInstance;
	}

	@Override
	public void onSource(INKFRequestContext aContext) throws Exception
	{	Object result=null;
		HttpClientContext state=HttpClientContext.create();
		Object stateref=null;
		Object cookieref=null;
		int statusCode=-1;
		HttpResponse response=null;
		IHDSNode responseHeaders=null;
		long expires=0;
		String mimetype=null;
		boolean cache=aContext.getThisRequest().argumentExists("cache");
		CloseableHttpClient client=null;
		
		//Get the configuration - first by push, then by pull
		IHttpClientConfig httpConfig;
		if(aContext.getThisRequest().argumentExists("config"))
		{	httpConfig=aContext.source("arg:config", IHttpClientConfig.class);
		}
		else
		{	httpConfig=aContext.source(HTTP_CONFIG_URI_DEFAULT, IHttpClientConfig.class);
		}
		
		
		String url;
		if(aContext.getThisRequest().argumentExists("simple"))
		{	//Set the url from simple format
			url=aContext.getThisRequest().getArgumentValue("simple");
			url=escapeURL(url);
			HttpGet get=new HttpGet(url);
			try
			{	client=getHttpClient(httpConfig, state, false, aContext, url);
	          	response=client.execute(get, state);
	          	statusCode=response.getStatusLine().getStatusCode();
	          	if(statusCode>=400)
	          	{	throw new Exception("HTTP Request Error :"+response.getStatusLine().getStatusCode());
	          	}
	          	try
	          	{
	          		result=getRepresentationFromHttpResponse(response);
	          	}
	          	catch(Exception e)
	          	{	get.abort();
	          		throw e;
	          	}
	          	responseHeaders=getResponseHeaders(response.getAllHeaders());
				mimetype=getContentType(response.getFirstHeader(CONTENT_TYPE_HEADER));
				expires=getExpires(response);
			}
			finally
			{	get.abort();
			}
			if(aContext.getThisRequest().getRepresentationClass()==HttpClientResponseRepresentation.class)
			{	result = new HttpClientResponseRepresentation(response, result);
			}
			INKFResponse resp= aContext.createResponseFrom(result);
			resp.setHeader(HTTP_ACCESSOR_STATUS_CODE_METADATA, statusCode);
			resp.setHeader(HTTP_ACCESSOR_RESPONSE_HEADERS_METADATA,responseHeaders);
			resp.setMimeType(mimetype);
			if(0 < expires)
			{	resp.setExpiry(INKFResponse.EXPIRY_CONSTANT, expires);	//Expire at an absolute time
			}
			else
			{	resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
			}
		}
		else
		{	//Set the url
			url=aContext.getThisRequest().getArgumentValue("url");
			url=escapeURL(url);

			//Get state if there is any
			String stateuri=aContext.getThisRequest().getArgumentValue("state");
			if(stateuri!=null)
			{	if(!stateuri.startsWith(HttpStateAccessor.ACTIVE_HTTP_STATE))
				{ stateuri=aContext.source("arg:state",String.class);
				}
				HttpStateRepresentation sa=aContext.source(stateuri, HttpStateRepresentation.class);
				state=sa.getState();
			}
			stateref=state.getCookieStore();
			if(stateref!=null)
			{	cookieref=((CookieStore)stateref).getCookies();				
			}
			
			//Oauth credentials?
			OAuthCredentials oauth=null;
			if(aContext.getThisRequest().argumentExists("oauth"))
			{	oauth=aContext.source("arg:oauth", OAuthCredentials.class);				
			}
			
			//User Specified Headers
			String headeruri=aContext.getThisRequest().getArgumentValue("headers");
			IHDSNode headers=null;
			if(headeruri!=null)
			{	headers=aContext.source("arg:headers", IHDSNode.class);
			}
					
			Integer mode=(Integer)mModes.get(aContext.getThisRequest().getArgumentValue("activeType"));
			switch (mode.intValue())
			{	case HTTP_METHOD_GET_INT:
					boolean remote=true;
					boolean shouldCache=false;
					HttpCacheRepresentation cacheRep=null;
					if(cache)
					{	//Source from cache
						INKFRequest req=aContext.createRequest(aContext.getThisRequest().getArgumentValue("cache"));
						req.addArgument("url", url);
						req.setRepresentationClass(HttpCacheRepresentation.class);
						cacheRep=(HttpCacheRepresentation)aContext.issueRequest(req);
						/*
						 * Should be test for expired
						 * 
						if(cacheRep!=null)
						{	remote=false;
							statusCode=HttpStatus.SC_OK;
							mimetype=cacheRep.getMimeType();
							result=cacheRep;
						}
						*/
					}
					if(remote)
					{	HttpGet get=new HttpGet(url);
						if(headers!=null)
						{	this.setRequestHeaders(get, headers);
						}
						if(cacheRep!=null)
						{	get.setHeader("If-None-Match", cacheRep.getTag());
						}
						try
						{	client=getHttpClient(httpConfig, state, false, aContext, url);
							if(oauth!=null)
							{	oauthSignRequest(get, oauth);								
							}
		                  	response=client.execute(get, state);
		                  	statusCode=response.getStatusLine().getStatusCode();
		                  	if(statusCode==HttpStatus.SC_NOT_MODIFIED)
		                  	{	result=cacheRep;
		                  	}
		                  	else
		                  	{	
		                  		try
		                  		{
		                  			result=getRepresentationFromHttpResponse(response);
		                  		}
			                  	catch(Exception e)
			                  	{	get.abort();
			                  		throw e;
			                  	}
		                  		shouldCache=true;
		                  	}
		                  	responseHeaders=getResponseHeaders(response.getAllHeaders());
							mimetype=getContentType(response.getFirstHeader(CONTENT_TYPE_HEADER));
							expires=getExpires(response);
						}
						finally
						{	get.abort();
						}
						//Sink to cache
						if(cache && shouldCache)
						{	String etag=(String)responseHeaders.getFirstValue("ETag");
							if(etag!=null)
							{	INKFRequest req=aContext.createRequest(aContext.getThisRequest().getArgumentValue("cache"));
								req.addArgument("url", url);
								req.addPrimaryArgument(new HttpCacheRepresentation(etag, mimetype, 0, (ByteArrayRepresentation)result ));
								req.setVerb(INKFRequestReadOnly.VERB_SINK);
								aContext.issueRequest(req);
							}
						}
					}
				break;
				
				case HTTP_METHOD_POST_INT:
					HttpPost post=new HttpPost(url);
					post.getParams().setParameter("http.protocol.expect-continue", Boolean.TRUE);
					if(headers!=null)
					{	this.setRequestHeaders(post, headers);
					}
					if(aContext.getThisRequest().argumentExists("body")) 
					{	//Currently we're only supporting single body argument
						INKFResponseReadOnly<IReadableBinaryStreamRepresentation> resp=aContext.sourceForResponse("arg:body", IReadableBinaryStreamRepresentation.class);
						IReadableBinaryStreamRepresentation rep=resp.getRepresentation();
						post.setEntity(new ReadableBinaryRequestEntity(rep, resp.getMimeType()));
					}
					else if(aContext.getThisRequest().argumentExists("nvp"))
					{	IHDSNode nvpa=(IHDSNode)aContext.source("arg:nvp",IHDSNode.class);
						List<NameValuePair> formparams = new ArrayList<NameValuePair>();
						for(IHDSNode n: nvpa.getChildren())
						{	formparams.add(new BasicNameValuePair(n.getName(), (String)n.getValue()));
						}
						UrlEncodedFormEntity entity = new UrlEncodedFormEntity(formparams, "UTF-8");
						post.setEntity(entity);
					}
					try
					{	client=getHttpClient(httpConfig,state, true, aContext, url);
						if(oauth!=null)
						{	oauthSignRequest(post, oauth);								
						}
	                  	response=client.execute(post, state);
	                  	statusCode=response.getStatusLine().getStatusCode();
	                  	try
	                  	{
	                  		result=getRepresentationFromHttpResponse(response);
	                  	}
	                  	catch(Exception e)
	                  	{	post.abort();
	                  		throw e;
	                  	}
						responseHeaders=getResponseHeaders(response.getAllHeaders());
						mimetype=getContentType(response.getFirstHeader(CONTENT_TYPE_HEADER));
						expires=getExpires(response);
					}
					finally
					{	post.abort();
					}
				break;
				
				case HTTP_METHOD_PUT_INT:
				{
					HttpPut put=new HttpPut(url);
					put.getParams().setParameter("http.protocol.expect-continue", Boolean.TRUE);
					if(headers!=null)
					{	this.setRequestHeaders(put, headers);
					}
					INKFResponseReadOnly resp=aContext.sourceForResponse("arg:body", IReadableBinaryStreamRepresentation.class);
					IReadableBinaryStreamRepresentation rep=(IReadableBinaryStreamRepresentation)resp.getRepresentation();
					put.setEntity(new ReadableBinaryRequestEntity(rep, resp.getMimeType()));
					try
					{	client=getHttpClient(httpConfig, state, true, aContext, url);
						if(oauth!=null)
						{	oauthSignRequest(put, oauth);								
						}
	                 	response=client.execute(put, state);
	                 	statusCode=response.getStatusLine().getStatusCode();
	                 	try
	                 	{
	                 		result=getRepresentationFromHttpResponse(response);
	                 	}
						catch(Exception e)
	                  	{	put.abort();
	                  		throw e;
	                  	}
						responseHeaders=getResponseHeaders(response.getAllHeaders());
						mimetype=getContentType(response.getFirstHeader(CONTENT_TYPE_HEADER));
						expires=getExpires(response);
					}
					finally
					{	put.abort();						
					}
				}
				break;
				case HTTP_METHOD_PATCH_INT:
				{
					HttpPatch patch=new HttpPatch(url);
					patch.getParams().setParameter("http.protocol.expect-continue", Boolean.TRUE);
					if(headers!=null)
					{	this.setRequestHeaders(patch, headers);
					}
					INKFResponseReadOnly resp=aContext.sourceForResponse("arg:body", IReadableBinaryStreamRepresentation.class);
					IReadableBinaryStreamRepresentation rep=(IReadableBinaryStreamRepresentation)resp.getRepresentation();
					patch.setEntity(new ReadableBinaryRequestEntity(rep, resp.getMimeType()));
					try
					{	client=getHttpClient(httpConfig, state, true, aContext, url);
						if(oauth!=null)
						{	oauthSignRequest(patch, oauth);								
						}
	                 	response=client.execute(patch, state);
	                 	statusCode=response.getStatusLine().getStatusCode();
	                 	try
	                 	{
	                 		result=getRepresentationFromHttpResponse(response);
						}
	                  	catch(Exception e)
	                  	{	patch.abort();
	                  		throw e;
	                  	}
						responseHeaders=getResponseHeaders(response.getAllHeaders());
						mimetype=getContentType(response.getFirstHeader(CONTENT_TYPE_HEADER));
						expires=getExpires(response);
					}
					finally
					{	patch.abort();						
					}
				}
				break;
				
				case HTTP_METHOD_HEAD_INT:
					HttpHead head=new HttpHead(url);
					if(headers!=null)
					{	this.setRequestHeaders(head, headers);
					}
					try
					{	client=getHttpClient(httpConfig, state, false, aContext, url);
						if(oauth!=null)
						{	oauthSignRequest(head, oauth);								
						}
	                  	response=client.execute(head, state);
	                  	statusCode=response.getStatusLine().getStatusCode();
	                  	responseHeaders=getResponseHeaders(response.getAllHeaders());
	                  	result=responseHeaders;
						mimetype=getContentType(response.getFirstHeader(CONTENT_TYPE_HEADER));
						expires=getExpires(response);
					}
					finally
					{	head.abort();						
					}
				break;
				case HTTP_METHOD_OPTIONS_INT:
					HttpOptions options=new HttpOptions(url);
					if(headers!=null)
					{	this.setRequestHeaders(options, headers);
					}
					try
					{	client=getHttpClient(httpConfig, state, false, aContext, url);
						if(oauth!=null)
						{	oauthSignRequest(options, oauth);								
						}
	                  	response=client.execute(options, state);
	                  	statusCode=response.getStatusLine().getStatusCode();
	                  	responseHeaders=getResponseHeaders(response.getAllHeaders());
	                  	result=getRepresentationFromHttpResponse(response);
						mimetype=getContentType(response.getFirstHeader(CONTENT_TYPE_HEADER));
						expires=getExpires(response);
					}
					finally
					{	options.abort();
					}
				break;
				case HTTP_METHOD_DELETE_INT:
					HttpDelete delete=new HttpDelete(url);
					if(headers!=null)
					{	this.setRequestHeaders(delete, headers);
					}
					try
					{	client=getHttpClient(httpConfig, state, false, aContext, url);
						if(oauth!=null)
						{	oauthSignRequest(delete, oauth);								
						}
	                  	response=client.execute(delete, state);
	                  	statusCode=response.getStatusLine().getStatusCode();
	                  	try
	                  	{	result=getRepresentationFromHttpResponse(response);
	                  	}
	                  	catch(Exception e)
	                  	{	delete.abort();
	                  		throw e;
	                  	}
	                  	responseHeaders=getResponseHeaders(response.getAllHeaders());
						mimetype=getContentType(response.getFirstHeader(CONTENT_TYPE_HEADER));
						expires=getExpires(response);
					}
					finally
					{	delete.abort();						
					}
					
				break;
			}
			
			//Consume content so that connection is released to manager.
			if(response!=null)
			{	HttpEntity entity=response.getEntity();
				if(entity!=null)
				{	EntityUtils.consume(entity);
				}
			}
			
			//Sink updated state
			CookieStore cs=state.getCookieStore();
			if(cs!=null && cookieref!=null && !cookieref.equals(cs.getCookies()))				
			{	HttpStateRepresentation sa=new HttpStateRepresentation(state);
				aContext.sink(stateuri, sa);
			}
			if(aContext.getThisRequest().getRepresentationClass()==HttpClientResponseRepresentation.class)
			{	result = new HttpClientResponseRepresentation(response, result);
			}
			INKFResponse resp= aContext.createResponseFrom(result);
			resp.setHeader(HTTP_ACCESSOR_STATUS_CODE_METADATA, statusCode);
			resp.setHeader(HTTP_ACCESSOR_RESPONSE_HEADERS_METADATA,responseHeaders);
			resp.setMimeType(mimetype);
			//resp.setNoCache();
			if(expires>0)
			{	resp.setExpiry(INKFResponse.EXPIRY_CONSTANT, expires);	//Expire at an absolute time
			}
			else
			{	resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
			}

		}
	}
	
	protected CloseableHttpClient getHttpClient(IHttpClientConfig cfg, HttpClientContext state, boolean entityEnclosingMethod, INKFRequestContext aContext, String aURL) throws Exception
	{	
		RequestConfig.Builder rqb=getAndIntitiallyConfigureRequestConfigBuilder(cfg, state, entityEnclosingMethod, aContext, aURL);
		
		//Source Connection Manager
		HttpClientConnectionManager cm=null;
		if(aContext.getThisRequest().argumentExists("connectionManager"))
		{	cm=aContext.source("arg:connectionManager", HttpClientConnectionManager.class);			
		}
		else
		{	cm=aContext.source("active:httpClientDefaultConnectionManager", HttpClientConnectionManager.class);
		}
		
		//Prepare to Instantiate Client
		HttpClientBuilder cb=HttpClientBuilder.create()
						.setConnectionManager(cm)
						.setRetryHandler(new DefaultHttpRequestRetryHandler(cfg.getNumberOfRetryAttempts(), false))
						.setUserAgent(cfg.getUserAgent())
						.setDefaultCredentialsProvider(state.getCredentialsProvider())
						;
		
		
		String proxyHost=cfg.getProxyHost();
		if(proxyHost!=null)
		{	URI url=new URI(aURL);
			String host=url.getHost();
			if(!cfg.isProxyingExcluded(host))
			{
				HttpHost proxy = new HttpHost(proxyHost, cfg.getProxyPort());
				//client.getParams().setParameter(ConnRoutePNames.DEFAULT_PROXY, proxy);
				cb.setProxy(proxy);
			
				String user=cfg.getProxyUser();
				String password=cfg.getProxyPassword();
				Credentials cred=null;
				if (cfg.getProxyNTLMDomain()!=null)
				{	cred = new NTCredentials(user, password, cfg.getProxyNTLMHost(), cfg.getProxyNTLMDomain());
					//client.getAuthSchemes().register("ntlm", new NTLMSchemeFactory());  - legacy used with the JCIFS library
					List<String> authpref = new ArrayList<String>();
					authpref.add(AuthSchemes.NTLM);
					//client.getParams().setParameter(AuthPNames.PROXY_AUTH_PREF, authpref);
					rqb.setProxyPreferredAuthSchemes(authpref);
					
					//RequestConfig.Builder.setProxyPreferredAuthSchemes
					aContext.logRaw(INKFRequestContext.LEVEL_DEBUG, "Registered NTLM proxy route and set PROXY_AUTH_PREF to NTLM" );
				}
				else if (user!=null && password!=null)
				{	cred=new UsernamePasswordCredentials(user,password);
				}
				if(cred!=null)
				{	CredentialsProvider cp=state.getCredentialsProvider();
					if(cp==null)
					{	cp=new BasicCredentialsProvider();
						state.setCredentialsProvider(cp);
					}
					AuthScope scope=new AuthScope(proxyHost, cfg.getProxyPort());
					cp.setCredentials(scope, cred);
				}
			}
		}
		rqb.setTargetPreferredAuthSchemes(cfg.getAuthPreferenceOrder());
		
		cb.setDefaultRequestConfig(rqb.build());
		
		return cb.build();
	}
	
}
