package org.netkernel.client.http.endpoint;

import java.net.URI;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;
import java.util.concurrent.TimeUnit;

import org.apache.commons.httpclient.methods.HttpPatch;
import org.apache.http.HttpEntity;
import org.apache.http.HttpHost;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.NameValuePair;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.Credentials;
import org.apache.http.auth.NTCredentials;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.CookieStore;
import org.apache.http.client.CredentialsProvider;
import org.apache.http.client.config.AuthSchemes;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpHead;
import org.apache.http.client.methods.HttpOptions;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.protocol.HttpClientContext;
import org.apache.http.concurrent.FutureCallback;
import org.apache.http.cookie.Cookie;
import org.apache.http.impl.client.BasicCredentialsProvider;
import org.apache.http.impl.nio.client.CloseableHttpAsyncClient;
import org.apache.http.impl.nio.client.HttpAsyncClientBuilder;
import org.apache.http.impl.nio.conn.PoolingNHttpClientConnectionManager;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.nio.conn.NHttpClientConnectionManager;
import org.apache.http.util.EntityUtils;
import org.netkernel.client.http.representation.HttpCacheRepresentation;
import org.netkernel.client.http.representation.HttpClientResponseRepresentation;
import org.netkernel.client.http.representation.HttpStateRepresentation;
import org.netkernel.client.http.representation.IHttpClientConfig;
import org.netkernel.client.http.representation.OAuthCredentials;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.nkf.impl.NKFContextImpl;
import org.netkernel.layer0.nkf.impl.NKFResponseImpl;
import org.netkernel.layer0.representation.ByteArrayRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.request.IResponse;

public class HttpClientAsyncAccessorImpl extends HttpClientAccessorBase
{	
	private static HttpClientAsyncAccessorImpl sInstance;
	
	private Map<IHttpClientConfig, SelfCleaningRunningClient> mClientMap = 
			java.util.Collections.synchronizedMap(new WeakHashMap<IHttpClientConfig, SelfCleaningRunningClient>());
	
	/** Creates a new instance of HTTPClientAccessor */
	public HttpClientAsyncAccessorImpl()
	{	super();
		sInstance=this;
	}
	
	public static HttpClientAsyncAccessorImpl getInstance()
	{	return sInstance;
	}
	
	@Override
	public void onDecommissionEndpoint() throws Exception
	{	Iterator it=mClientMap.keySet().iterator();
		while(it.hasNext())
		{	SelfCleaningRunningClient cleanableClient=mClientMap.get(it.next());
			CloseableHttpAsyncClient client=cleanableClient.getRunningClient();
			try
			{	client.close();
			}
			catch(Exception e)
			{	e.printStackTrace();
			}
		}
		mClientMap.clear();
		super.onDecommissionEndpoint();
	}

	@Override
	public void onSource(INKFRequestContext aContext) throws Exception
	{	HttpClientContext httpClientContext=HttpClientContext.create();
		Object stateref=null;
		List<Cookie> cookieref=null;
		boolean cache=aContext.getThisRequest().argumentExists("cache");
		
		//Get the configuration - first by push, then by pull
		IHttpClientConfig httpConfig;
		if(aContext.getThisRequest().argumentExists("config"))
		{	httpConfig=aContext.source("arg:config", IHttpClientConfig.class);
		}
		else
		{	httpConfig=aContext.source(HTTP_CONFIG_URI_DEFAULT, IHttpClientConfig.class);
		}
		
		String url;
		{	//Set the url
			url=aContext.getThisRequest().getArgumentValue("url");
			url=escapeURL(url);

			//Get state if there is any
			String stateuri=aContext.getThisRequest().getArgumentValue("state");
			if(stateuri!=null)
			{	if(!stateuri.startsWith(HttpStateAccessor.ACTIVE_HTTP_STATE))
				{ stateuri=aContext.source("arg:state",String.class);
				}
				HttpStateRepresentation sa=aContext.source(stateuri, HttpStateRepresentation.class);
				httpClientContext=sa.getState();
			}
			stateref=httpClientContext.getCookieStore();
			if(stateref!=null)
			{	cookieref=((CookieStore)stateref).getCookies();				
			}
			
			//Oauth credentials?
			OAuthCredentials oauth=null;
			if(aContext.getThisRequest().argumentExists("oauth"))
			{	oauth=aContext.source("arg:oauth", OAuthCredentials.class);				
			}
			
			//User Specified Headers
			String headeruri=aContext.getThisRequest().getArgumentValue("headers");
			IHDSNode headers=null;
			if(headeruri!=null)
			{	headers=aContext.source("arg:headers", IHDSNode.class);
			}
					
			Integer mode=(Integer)mModes.get(aContext.getThisRequest().getArgumentValue("activeType"));
			switch (mode.intValue())
			{	case HTTP_METHOD_GET_INT:
					boolean remote=true;
					HttpCacheRepresentation cacheRep=null;
					if(cache)
					{	//Source from cache
						INKFRequest req=aContext.createRequest(aContext.getThisRequest().getArgumentValue("cache"));
						req.addArgument("url", url);
						req.setRepresentationClass(HttpCacheRepresentation.class);
						cacheRep=(HttpCacheRepresentation)aContext.issueRequest(req);
						/*
						 * Should be test for expired
						 * 
						if(cacheRep!=null)
						{	remote=false;
							statusCode=HttpStatus.SC_OK;
							mimetype=cacheRep.getMimeType();
							result=cacheRep;
						}
						*/
					}
					if(remote)
					{	HttpGet get=new HttpGet(url);
					
						if(headers!=null)
						{	this.setRequestHeaders(get, headers);
						}
						if(cacheRep!=null)
						{	get.setHeader("If-None-Match", cacheRep.getTag());
						}
						try
						{	CloseableHttpAsyncClient client=getHttpClient(httpConfig, httpClientContext, false, aContext, url);
							if(oauth!=null)
							{	oauthSignRequest(get, oauth);								
							}
							String cacheurl=null;
							if(cache)
							{	cacheurl=url;
							}
		                  	client.execute(get, httpClientContext, new GetFutureCallbackImpl(aContext, get, client, httpClientContext, cookieref, stateuri, cacheurl, cacheRep));
						}
						finally
						{	//get.abort();
						}
					}
				break;
				
				
				case HTTP_METHOD_POST_INT:
					HttpPost post=new HttpPost(url);
					post.getParams().setParameter("http.protocol.expect-continue", Boolean.TRUE);
					if(headers!=null)
					{	this.setRequestHeaders(post, headers);
					}
					if(aContext.getThisRequest().argumentExists("body")) 
					{	//Currently we're only supporting single body argument
						INKFResponseReadOnly<IReadableBinaryStreamRepresentation> resp=aContext.sourceForResponse("arg:body", IReadableBinaryStreamRepresentation.class);
						IReadableBinaryStreamRepresentation rep=resp.getRepresentation();
						post.setEntity(new ReadableBinaryRequestEntity(rep, resp.getMimeType()));
					}
					else if(aContext.getThisRequest().argumentExists("nvp"))
					{	IHDSNode nvpa=(IHDSNode)aContext.source("arg:nvp",IHDSNode.class);
						List<NameValuePair> formparams = new ArrayList();
						for(IHDSNode n: nvpa.getChildren())
						{	formparams.add(new BasicNameValuePair(n.getName(), (String)n.getValue()));
						}
						UrlEncodedFormEntity entity = new UrlEncodedFormEntity(formparams, "UTF-8");
						post.setEntity(entity);
					}
					try
					{	CloseableHttpAsyncClient client=getHttpClient(httpConfig,httpClientContext, true, aContext, url);
						if(oauth!=null)
						{	oauthSignRequest(post, oauth);								
						}
	                  	client.execute(post, httpClientContext, new PostFutureCallbackImpl(aContext, post, client, httpClientContext, cookieref, stateuri));
					}
					finally
					{	//post.abort();
					}
				break;
				
				case HTTP_METHOD_PUT_INT:
				{
					HttpPut put=new HttpPut(url);
					put.getParams().setParameter("http.protocol.expect-continue", Boolean.TRUE);
					if(headers!=null)
					{	this.setRequestHeaders(put, headers);
					}
					INKFResponseReadOnly resp=aContext.sourceForResponse("arg:body", IReadableBinaryStreamRepresentation.class);
					IReadableBinaryStreamRepresentation rep=(IReadableBinaryStreamRepresentation)resp.getRepresentation();
					put.setEntity(new ReadableBinaryRequestEntity(rep, resp.getMimeType()));
					try
					{	CloseableHttpAsyncClient client=getHttpClient(httpConfig, httpClientContext, true, aContext, url);
						if(oauth!=null)
						{	oauthSignRequest(put, oauth);								
						}
	                 	client.execute(put, httpClientContext, new PutFutureCallbackImpl(aContext, put, client, httpClientContext, cookieref, stateuri));
					}
					finally
					{	//put.abort();						
					}
				}
				break;
				case HTTP_METHOD_PATCH_INT:
				{
					HttpPatch patch=new HttpPatch(url);
					patch.getParams().setParameter("http.protocol.expect-continue", Boolean.TRUE);
					if(headers!=null)
					{	this.setRequestHeaders(patch, headers);
					}
					INKFResponseReadOnly resp=aContext.sourceForResponse("arg:body", IReadableBinaryStreamRepresentation.class);
					IReadableBinaryStreamRepresentation rep=(IReadableBinaryStreamRepresentation)resp.getRepresentation();
					patch.setEntity(new ReadableBinaryRequestEntity(rep, resp.getMimeType()));
					try
					{	CloseableHttpAsyncClient client=getHttpClient(httpConfig, httpClientContext, true, aContext, url);
						if(oauth!=null)
						{	oauthSignRequest(patch, oauth);								
						}
	                 	client.execute(patch, httpClientContext, new PatchFutureCallbackImpl(aContext, patch, client, httpClientContext, cookieref, stateuri));
					}
					finally
					{	//patch.abort();						
					}
				}
				break;
				case HTTP_METHOD_HEAD_INT:
					HttpHead head=new HttpHead(url);
					if(headers!=null)
					{	this.setRequestHeaders(head, headers);
					}
					try
					{	CloseableHttpAsyncClient client=getHttpClient(httpConfig, httpClientContext, false, aContext, url);
						if(oauth!=null)
						{	oauthSignRequest(head, oauth);								
						}
	                  	client.execute(head, httpClientContext, new HeadFutureCallbackImpl(aContext, head, client, httpClientContext, cookieref, stateuri));
	                  	
					}
					finally
					{	//head.abort();						
					}
				break;
				case HTTP_METHOD_OPTIONS_INT:
					HttpOptions options=new HttpOptions(url);
					if(headers!=null)
					{	this.setRequestHeaders(options, headers);
					}
					try
					{	CloseableHttpAsyncClient client=getHttpClient(httpConfig, httpClientContext, false, aContext, url);
						if(oauth!=null)
						{	oauthSignRequest(options, oauth);								
						}
	                  	client.execute(options, httpClientContext, new OptionsFutureCallbackImpl(aContext, options, client, httpClientContext, cookieref, stateuri));
	                  	
					}
					finally
					{	//head.abort();						
					}
				break;
				
				case HTTP_METHOD_DELETE_INT:
					HttpDelete delete=new HttpDelete(url);
					if(headers!=null)
					{	this.setRequestHeaders(delete, headers);
					}
					try
					{	CloseableHttpAsyncClient client=getHttpClient(httpConfig, httpClientContext, false, aContext, url);
						if(oauth!=null)
						{	oauthSignRequest(delete, oauth);								
						}
	                  	client.execute(delete, httpClientContext, new DeleteFutureCallbackImpl(aContext, delete, client, httpClientContext, cookieref, stateuri));
					}
					finally
					{	//delete.abort();						
					}
					
				break;
				
			}

			//Return immediately with no response set.
			aContext.setNoResponse();
		}
	}
	
	protected synchronized CloseableHttpAsyncClient getHttpClient(IHttpClientConfig cfg, HttpClientContext state, boolean entityEnclosingMethod, INKFRequestContext aContext, String aURL) throws Exception
	{	
		SelfCleaningRunningClient cleanableClient=null;
		CloseableHttpAsyncClient client=null;
		cleanableClient=mClientMap.get(cfg);
		if(cleanableClient!=null)
		{	client=cleanableClient.getRunningClient();
			if(client.isRunning())
			{	return client;
			}
			else
			{	//This  is a dead client so remove from map and create new one
				mClientMap.remove(cfg);
				client=null;
			}
		}
		
		RequestConfig.Builder rqb=getAndIntitiallyConfigureRequestConfigBuilder(cfg, state, entityEnclosingMethod, aContext, aURL);
		
		//Source Connection Manager
		NHttpClientConnectionManager cm=null;
		if(aContext.getThisRequest().argumentExists("connectionManager"))
		{	cm=aContext.source("arg:connectionManager", NHttpClientConnectionManager.class);			
		}
		else
		{	INKFRequest req=aContext.createRequest("active:httpAsyncClientDefaultConnectionManager");
			req.addArgumentByValue("key", cfg);
			req.setRepresentationClass(PoolingNHttpClientConnectionManager.class);
			req.setHeader(INKFRequest.HEADER_EXCLUDE_DEPENDENCIES, true);		//This is not a dependency on the representation can exclude it.
			PoolingNHttpClientConnectionManager connMgr=(PoolingNHttpClientConnectionManager)aContext.issueRequest(req);
			// Close expired connections
			connMgr.closeExpiredConnections();
	        
	        // Close connections that have been idle longer than connection timeout
	        connMgr.closeIdleConnections(cfg.getConnectTimeout(), TimeUnit.MILLISECONDS);
			
			cm=connMgr;

		}
		
		//Prepare to Instantiate Client
		HttpAsyncClientBuilder cb=HttpAsyncClientBuilder.create()
						//.setConnectionManager(cm)
						//.setRetryHandler(new DefaultHttpRequestRetryHandler(cfg.getNumberOfRetryAttempts(), false))
						.setUserAgent(cfg.getUserAgent())
						;
		if(cm!=null)
		{	cb.setConnectionManager(cm);
		}
		
		String proxyHost=cfg.getProxyHost();
		if(proxyHost!=null)
		{	URI url=new URI(aURL);
			String host=url.getHost();
			if(!cfg.isProxyingExcluded(host))
			{
				HttpHost proxy = new HttpHost(proxyHost, cfg.getProxyPort());
				//client.getParams().setParameter(ConnRoutePNames.DEFAULT_PROXY, proxy);
				cb.setProxy(proxy);
			
				String user=cfg.getProxyUser();
				String password=cfg.getProxyPassword();
				Credentials cred=null;
				if (cfg.getProxyNTLMDomain()!=null)
				{	cred = new NTCredentials(user, password, cfg.getProxyNTLMHost(), cfg.getProxyNTLMDomain());
					//client.getAuthSchemes().register("ntlm", new NTLMSchemeFactory());  - legacy used with the JCIFS library
					List<String> authpref = new ArrayList<String>();
					authpref.add(AuthSchemes.NTLM);
					//client.getParams().setParameter(AuthPNames.PROXY_AUTH_PREF, authpref);
					rqb.setProxyPreferredAuthSchemes(authpref);
					
					//RequestConfig.Builder.setProxyPreferredAuthSchemes
					aContext.logRaw(INKFRequestContext.LEVEL_DEBUG, "Registered NTLM proxy route and set PROXY_AUTH_PREF to NTLM" );
				}
				else if (user!=null && password!=null)
				{	cred=new UsernamePasswordCredentials(user,password);
				}
				if(cred!=null)
				{	CredentialsProvider cp=state.getCredentialsProvider();
					if(cp==null)
					{	cp=new BasicCredentialsProvider();
						state.setCredentialsProvider(cp);
					}
					AuthScope scope=new AuthScope(proxyHost, cfg.getProxyPort());
					cp.setCredentials(scope, cred);
				}
			}
		}
		rqb.setTargetPreferredAuthSchemes(cfg.getAuthPreferenceOrder());
		
		cb.setDefaultRequestConfig(rqb.build());
		
		client=cb.build();
		client.start();
		mClientMap.put(cfg, new SelfCleaningRunningClient(client));
		return client;
	}
	
	
	class GetFutureCallbackImpl extends BaseFutureCallbackImpl
	{
		HttpGet mGet;
		String mCacheURL;
		HttpCacheRepresentation mCacheRep;
		public GetFutureCallbackImpl(INKFRequestContext aContext, HttpGet aGet, CloseableHttpAsyncClient aClient, HttpClientContext aState, List<Cookie> aCookieRef, String aStateURI, String cacheurl, HttpCacheRepresentation cacheRep)
		{	super(aContext, aClient, aState, aCookieRef, aStateURI);
			mGet=aGet;
			mCacheURL=cacheurl;
			mCacheRep=cacheRep;
		}

		@Override
		public void completed(HttpResponse response) {
			try {
				this.sinkCookieState();
				
				Object result=getRepresentationFromHttpResponse(response);
				if(response.getStatusLine().getStatusCode()==HttpStatus.SC_NOT_MODIFIED)
				{	result=mCacheRep;
				}
				else if (response.getStatusLine().getStatusCode()==HttpStatus.SC_OK && mCacheURL!=null) 
				{	//Sink to cache
					String etag=(String)getResponseHeaders(response.getAllHeaders()).getFirstValue("ETag");
					if(etag!=null)
					{	INKFRequest req=mContext.createRequest(mContext.getThisRequest().getArgumentValue("cache"));
						req.addArgument("url", mCacheURL);
						req.addPrimaryArgument(new HttpCacheRepresentation(etag, getResponseMimeType(response), 0, (ByteArrayRepresentation)result));
						req.setVerb(INKFRequestReadOnly.VERB_SINK);
						mContext.issueRequest(req);		//Can we abuse this mContext like this?
					}
				}
				
				INKFResponse resp = this.prepareNetKernelResponse(response, result, mContext);
				long expires=getExpires(response);
				if(expires>0)
				{	resp.setExpiry(INKFResponse.EXPIRY_CONSTANT, expires);	//Expire at an absolute time
				}
				else
				{	resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
				}
				this.issueAsyncNetKernelResponse(resp);
			}
			catch(Exception e) {
				this.handleAsyncException(e);
			}
			finally
			{	try
				{	this.gracefullyReleaseConnection(response);
				}
				catch(Exception e)
				{	e.printStackTrace();					
				}
				mGet.releaseConnection();
				mGet.completed();
			}
		}
	}
	
	class PostFutureCallbackImpl extends BaseFutureCallbackImpl
	{
		HttpPost mPost;
		public PostFutureCallbackImpl(INKFRequestContext aContext, HttpPost aPost, CloseableHttpAsyncClient aClient,  HttpClientContext aState, List<Cookie> aCookieRef, String aStateURI)
		{	super(aContext, aClient, aState, aCookieRef,aStateURI);
			mPost=aPost;
		}

		@Override
		public void completed(HttpResponse response) {	
			try {
				this.sinkCookieState();
				
				INKFResponse resp = this.prepareNetKernelResponse(response, getRepresentationFromHttpResponse(response), mContext);
				resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
				this.issueAsyncNetKernelResponse(resp);
			}
			catch(Exception e) {
				this.handleAsyncException(e);
			}
			finally
			{	try
				{	this.gracefullyReleaseConnection(response);
				}
				catch(Exception e)
				{	e.printStackTrace();					
				}
				mPost.releaseConnection();
				mPost.completed();
			}
		}
	}
	
	class PutFutureCallbackImpl extends BaseFutureCallbackImpl
	{
		HttpPut mPut;
		public PutFutureCallbackImpl(INKFRequestContext aContext, HttpPut aPut, CloseableHttpAsyncClient aClient,  HttpClientContext aState, List<Cookie> aCookieRef, String aStateURI)
		{	super(aContext, aClient, aState, aCookieRef, aStateURI);
			mPut=aPut;
		}

		@Override
		public void completed(HttpResponse response) {	
			try {
				this.sinkCookieState();
				
				INKFResponse resp = this.prepareNetKernelResponse(response, getRepresentationFromHttpResponse(response), mContext);
				resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
				this.issueAsyncNetKernelResponse(resp);
			}
			catch(Exception e) {
				this.handleAsyncException(e);
			}
			finally
			{	
				try
				{	this.gracefullyReleaseConnection(response);
				}
				catch(Exception e)
				{	e.printStackTrace();					
				}
				mPut.releaseConnection();
				mPut.completed();
			}
		}
	}
	
	class PatchFutureCallbackImpl extends BaseFutureCallbackImpl
	{
		HttpPatch mPatch;
		public PatchFutureCallbackImpl(INKFRequestContext aContext, HttpPatch aPatch, CloseableHttpAsyncClient aClient,  HttpClientContext aState, List<Cookie> aCookieRef, String aStateURI)
		{	super(aContext, aClient, aState, aCookieRef, aStateURI);
			mPatch=aPatch;
		}

		@Override
		public void completed(HttpResponse response) {	
			try {
				this.sinkCookieState();
				
				INKFResponse resp = this.prepareNetKernelResponse(response, getRepresentationFromHttpResponse(response), mContext);
				resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
				this.issueAsyncNetKernelResponse(resp);
			}
			catch(Exception e) {
				this.handleAsyncException(e);
			}
			finally
			{	
				try
				{	this.gracefullyReleaseConnection(response);
				}
				catch(Exception e)
				{	e.printStackTrace();					
				}
				mPatch.releaseConnection();
				mPatch.completed();
			}
		}
	}
	
	class HeadFutureCallbackImpl extends BaseFutureCallbackImpl
	{
		HttpHead mHead;
		public HeadFutureCallbackImpl(INKFRequestContext aContext, HttpHead aHead, CloseableHttpAsyncClient aClient,  HttpClientContext aState, List<Cookie> aCookieRef, String aStateURI)
		{	super(aContext, aClient, aState, aCookieRef, aStateURI);
			mHead=aHead;
		}

		@Override
		public void completed(HttpResponse response) {	
			try {
				this.sinkCookieState();
				
				INKFResponse resp = this.prepareNetKernelResponse(response, getResponseHeaders(response.getAllHeaders()), mContext);
				long expires=getExpires(response);
				if(expires>0)
				{	resp.setExpiry(INKFResponse.EXPIRY_CONSTANT, expires);	//Expire at an absolute time
				}
				else
				{	resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
				}
				this.issueAsyncNetKernelResponse(resp);
			}
			catch(Exception e) {
				this.handleAsyncException(e);
			}
			finally
			{	
				try
				{	this.gracefullyReleaseConnection(response);
				}
				catch(Exception e)
				{	e.printStackTrace();					
				}
				mHead.releaseConnection();
				mHead.completed();
			}
		}
	}
	
	class OptionsFutureCallbackImpl extends BaseFutureCallbackImpl
	{
		HttpOptions mOptions;
		public OptionsFutureCallbackImpl(INKFRequestContext aContext, HttpOptions aOptions, CloseableHttpAsyncClient aClient,  HttpClientContext aState, List<Cookie> aCookieRef, String aStateURI)
		{	super(aContext, aClient, aState, aCookieRef, aStateURI);
			mOptions=aOptions;
		}

		@Override
		public void completed(HttpResponse response) {	
			try {
				this.sinkCookieState();
				Object result=getRepresentationFromHttpResponse(response);
				INKFResponse resp = this.prepareNetKernelResponse(response, result, mContext);
				long expires=getExpires(response);
				if(expires>0)
				{	resp.setExpiry(INKFResponse.EXPIRY_CONSTANT, expires);	//Expire at an absolute time
				}
				else
				{	resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
				}
				this.issueAsyncNetKernelResponse(resp);
			}
			catch(Exception e) {
				this.handleAsyncException(e);
			}
			finally
			{	
				try
				{	this.gracefullyReleaseConnection(response);
				}
				catch(Exception e)
				{	e.printStackTrace();					
				}
				mOptions.releaseConnection();
				mOptions.completed();
			}
		}
	}
	
	class DeleteFutureCallbackImpl extends BaseFutureCallbackImpl
	{
		HttpDelete mDelete;
		public DeleteFutureCallbackImpl(INKFRequestContext aContext, HttpDelete aDelete, CloseableHttpAsyncClient aClient,  HttpClientContext aState, List<Cookie> aCookieRef, String aStateURI)
		{	super(aContext, aClient,  aState, aCookieRef, aStateURI);
			mDelete=aDelete;
		}

		@Override
		public void completed(HttpResponse response) {
			try {
				this.sinkCookieState();
				
				INKFResponse resp = this.prepareNetKernelResponse(response, getRepresentationFromHttpResponse(response), mContext);
				resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
				this.issueAsyncNetKernelResponse(resp);
			}
			catch(Exception e) {
				this.handleAsyncException(e);
			}
			finally
			{	
				try
				{	this.gracefullyReleaseConnection(response);
				}
				catch(Exception e)
				{	e.printStackTrace();					
				}
				mDelete.releaseConnection();
				mDelete.completed();
			}
		}
	}
	
	class BaseFutureCallbackImpl implements FutureCallback<HttpResponse> {
		protected INKFRequestContext mContext;
		protected CloseableHttpAsyncClient mClient;
		protected HttpClientContext mState;
		protected List<Cookie> mCookieRef;
		protected String mStateURI;
		
		public BaseFutureCallbackImpl(INKFRequestContext aContext, CloseableHttpAsyncClient aClient, HttpClientContext aState, List<Cookie> aCookieRef, String aStateURI) {	
			mContext = aContext;
			mClient=aClient;
			mState=aState;
			mCookieRef=aCookieRef;
			mStateURI=aStateURI;
		}
		
		public void sinkCookieState() throws Exception
		{	CookieStore cs=mState.getCookieStore();
			if(cs!=null && mCookieRef!=null && !mCookieRef.equals(cs.getCookies()))				
			{	HttpStateRepresentation sa=new HttpStateRepresentation(mState);
				mContext.sink(mStateURI, sa);		//TODO - check that this is correct - this context has had setNoResponse() !!
			}
		}

		@Override
		public void cancelled() {	
			NKFException e = new NKFException("HTTP Request Cancelled");
			INKFResponse vResponse = mContext.createResponseFrom(e);
			IResponse vKernelResponse = ((NKFResponseImpl)vResponse).getKernelResponse();
			((NKFContextImpl)mContext).handleAsyncResponse(vKernelResponse);
		}

		@Override
		public void completed(HttpResponse response) {	
			//Not implemented in base class
		}

		@Override
		public void failed(Exception e) {	
			INKFResponse vResponse = mContext.createResponseFrom(e);
			IResponse vKernelResponse = ((NKFResponseImpl)vResponse).getKernelResponse();
			((NKFContextImpl)mContext).handleAsyncResponse(vKernelResponse);
		}
		
		public void gracefullyReleaseConnection(HttpResponse response) throws Exception
		{	//Don't close the client as that shuts down the connection pool!
			//mClient.close();
			if(response!=null)
			{	HttpEntity entity=response.getEntity();
				if(entity!=null)
				{	EntityUtils.consume(entity);
				}
			}
		}
		
		public INKFResponse prepareNetKernelResponse(HttpResponse response, Object result, INKFRequestContext aContext) throws Exception
		{	if(aContext.getThisRequest().getRepresentationClass()==HttpClientResponseRepresentation.class)
			{	result = new HttpClientResponseRepresentation(response, result);
			}
			INKFResponse resp= mContext.createResponseFrom(result);
			resp.setHeader(HTTP_ACCESSOR_STATUS_CODE_METADATA, response.getStatusLine().getStatusCode());
			resp.setHeader(HTTP_ACCESSOR_RESPONSE_HEADERS_METADATA,getResponseHeaders(response.getAllHeaders()));
			resp.setMimeType(getResponseMimeType(response));
			return resp;
		}
		
		public String getResponseMimeType(HttpResponse response)
		{	return getContentType(response.getFirstHeader(CONTENT_TYPE_HEADER));
		}
		
		public void issueAsyncNetKernelResponse(INKFResponse resp)
		{	IResponse vKernelResponse = ((NKFResponseImpl)resp).getKernelResponse();
			((NKFContextImpl)mContext).handleAsyncResponse(vKernelResponse);
		}
		
		public void handleAsyncException(Exception e)
		{	INKFResponse vResponse = mContext.createResponseFrom(e);
			IResponse vKernelResponse = ((NKFResponseImpl)vResponse).getKernelResponse();
			((NKFContextImpl)mContext).handleAsyncResponse(vKernelResponse);			
		}
	}
	
	private class SelfCleaningRunningClient
	{
		private CloseableHttpAsyncClient mClient;
		
		public SelfCleaningRunningClient(CloseableHttpAsyncClient aClient)
		{	mClient=aClient;			
		}
		
		public CloseableHttpAsyncClient getRunningClient()
		{	return mClient;			
		}

		@Override
		protected void finalize() throws Throwable
		{	mClient.close();
		}
	}
	 
}
