package org.netkernel.client.http.endpoint;


import org.apache.http.auth.Credentials;
import org.apache.http.auth.NTCredentials;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.netkernel.client.http.representation.HttpCredentialsRepresentation;
import org.netkernel.client.http.representation.IHttpCredentials;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.module.standard.endpoint.StandardTransreptorImpl;

public class HttpCredentialsTransreptor extends StandardTransreptorImpl
{
	public HttpCredentialsTransreptor()
	{	this.declareThreadSafe();
		this.declareToRepresentation(IHttpCredentials.class);
	}

	@Override
	public void onTransrept(INKFRequestContext aContext) throws Exception
	{	IHDSNode node=(IHDSNode)aContext.sourcePrimary(IHDSNode.class);
		String realm=null;
		String temp=(String)node.getFirstValue("/httpCredentials/realm");
		if(temp!=null)
		{	realm=temp;
		}
		String host=(String)node.getFirstValue("/httpCredentials/host");
		String username=(String)node.getFirstValue("/httpCredentials/username");
		String password=(String)node.getFirstValue("/httpCredentials/password");
		int port=80;
		temp=(String)node.getFirstValue("/httpCredentials/port");
		if(temp!=null)
		{	port=Integer.parseInt(temp);			
		}
		String nthost=null;
		temp=(String)node.getFirstValue("/httpCredentials/NTRequestHost");
		if(temp!=null)
		{	nthost=temp;
		}
		String ntdomain=null;
		temp=(String)node.getFirstValue("/httpCredentials/NTDomain");
		if(temp!=null)
		{	ntdomain=temp;
		}
		Credentials cred=null;
		if(ntdomain!=null && ntdomain!=null)
		{	cred=new NTCredentials(username, password, nthost, ntdomain);
		}
		else
		{	cred=new UsernamePasswordCredentials(username, password);
		}
		HttpCredentialsRepresentation ca=new HttpCredentialsRepresentation(realm, host, port, cred);
		
		aContext.createResponseFrom(ca);
	}	
	

}
