package org.netkernel.client.http.representation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.http.client.config.AuthSchemes;
import org.apache.http.client.params.AuthPolicy;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;

public class HttpClientConfig implements IHttpClientConfig
{
	private boolean mFollowRedirects;
	private long mMaximumAcceptableContentLength;
	private int mMaximumConnectionsPerHost;
	private int mTotalConnections;
	private int mStateExpirationTime;
	private int mRetryAttempts;
	private int mConnectTimeout;
	private int mTimeout;
	private String mProxyHost;
	private int mProxyPort;
	private String mProxyUser;
	private String mProxyPassword;
	private String mProxyNTLMHost;
	private String mProxyNTLMDomain;
	private boolean mExpectContinue;
	private String mUserAgent;
	List<String> mAuthPreferenceOrder = new ArrayList<String>();
	HashMap<String, Boolean> mProxyExcludes;
	ArrayList<String> mProxyExcludesWildcardStems;
	
	private IHDSNode mEqualityHDS;

/** Creates a new instance of HttpClientConfigAspect */
	public HttpClientConfig(
		boolean	aFollowRedirects,
		long aAcceptableContentLength,
		int aConnectionsPerHost,
		int aTotalConnections,
		int aStateExpirationTime,
		int aRetryAttempts,
		int aConnectTimeout,
		int aTimeout,
		String aProxyHost,
		int aProxyPort,
		String aProxyUser,
		String aProxyPassword,
		String aProxyNTLMHost,
		String aProxyNTLMDomain,
		boolean aExpectContinue,
		String aUserAgent,
		String aAuthPreferenceOrder,
		String aProxyExclude
	)
	{	mFollowRedirects=aFollowRedirects;
		mMaximumAcceptableContentLength =aAcceptableContentLength;
		mMaximumConnectionsPerHost =aConnectionsPerHost;
		mTotalConnections=aTotalConnections;
		mStateExpirationTime=aStateExpirationTime;
		mRetryAttempts=aRetryAttempts;
		mConnectTimeout=aConnectTimeout;
		mTimeout=aTimeout;
		mProxyHost=aProxyHost;
		mProxyPort=aProxyPort;
		mProxyUser=aProxyUser;
		mProxyPassword=aProxyPassword;
		mProxyNTLMHost=aProxyNTLMHost;
		mProxyNTLMDomain=aProxyNTLMDomain;
		mExpectContinue=aExpectContinue;
		mUserAgent=aUserAgent;
		if(aAuthPreferenceOrder==null)
		{	mAuthPreferenceOrder.add(AuthSchemes.DIGEST);
			mAuthPreferenceOrder.add(AuthSchemes.BASIC);
			//System.out.println(AuthSchemes.BASIC+":"+AuthSchemes.DIGEST+":"+AuthSchemes.NTLM);
		}
		else
		{	String[] sa=aAuthPreferenceOrder.split(",");
			for(int i=0; i< sa.length; i++)
			{	String s=sa[i].trim();
				mAuthPreferenceOrder.add(s);
			}
		}
		mProxyExcludes=new HashMap<String, Boolean>();
		mProxyExcludesWildcardStems=new ArrayList<String>();
		if(aProxyExclude!=null)
		{	String[] sa=aProxyExclude.split(",");
			for(int i=0; i< sa.length; i++)
			{	String s=sa[i].trim();
				//s=s.replaceAll("\\*\\.", "");
				if(s.startsWith("*"))
				{	mProxyExcludesWildcardStems.add(new StringBuilder(s.substring(1, s.length())).reverse().toString());										
				}
				else
				{	mProxyExcludes.put(s, true);
				}
			}
		}
		
		//Use HDS to generate unique hashcode
		HDSBuilder b=new HDSBuilder();
		b.addNode("a", mFollowRedirects);
		b.addNode("a", mMaximumAcceptableContentLength);
		b.addNode("a", mMaximumConnectionsPerHost);
		b.addNode("a", mStateExpirationTime);
		b.addNode("a", mRetryAttempts);
		b.addNode("a", mConnectTimeout);
		b.addNode("a", mTimeout);
		b.addNode("a", mProxyHost);
		b.addNode("a", mProxyPort);
		b.addNode("a", mProxyPassword);
		b.addNode("a", mProxyUser);
		b.addNode("a", mProxyNTLMHost);
		b.addNode("a", mProxyNTLMDomain);
		b.addNode("a", mExpectContinue);
		b.addNode("a", mUserAgent);
		mEqualityHDS=b.getRoot();
	}
	
	public boolean followRedirects()
	{	return mFollowRedirects;
	}
	
	public long getMaximumAccepatbleContentLength()
	{	return mMaximumAcceptableContentLength;
	}
	
	public int getMaximumConnectionsPerHost()
	{	return mMaximumConnectionsPerHost;
	}
	
	public int getMaximumTotalConnections()
	{	return mTotalConnections;
	}
	
	public int getStateExpirationTime()
	{	return mStateExpirationTime;
	}
	
	public int getNumberOfRetryAttempts()
	{	return mRetryAttempts;
	}
	
	public int getConnectTimeout()
	{	return mConnectTimeout;		
	}
	
	public int getTimeout()
	{	return mTimeout;		
	}
	
	public String getProxyHost()
	{	return mProxyHost;		
	}
	public int getProxyPort()
	{	return mProxyPort;		
	}
	public String getProxyUser()
	{	return mProxyUser;		
	}
	public String getProxyPassword()
	{	return mProxyPassword;		
	}
	public String getProxyNTLMHost()
	{	return mProxyNTLMHost;		
	}
	public String getProxyNTLMDomain()
	{	return mProxyNTLMDomain;		
	}
	public boolean expectContinue()
	{	return mExpectContinue;		
	}
	public String getUserAgent()
	{	return mUserAgent;		
	}
	
	public List<String> getAuthPreferenceOrder()
	{	return mAuthPreferenceOrder;
	}
	
	public boolean isProxyingExcluded(String aHostname)
	{	boolean b= mProxyExcludes.containsKey(aHostname);
		if(b==false)
		{	aHostname=new StringBuilder(aHostname).reverse().toString();
			for(String stem : mProxyExcludesWildcardStems)
			{	//System.out.println("STEM: "+stem);
				if(aHostname.startsWith(stem))
				{	b=true;
					break;
				}
				
			}
		}
		//System.out.println("PROXY: "+b);
		return b;
	}
	
	public boolean equals(Object other)
	{	return (other instanceof HttpClientConfig) && ((HttpClientConfig)other).mEqualityHDS.equals(mEqualityHDS);
	}
	
	public int hashCode()
	{	return mEqualityHDS.hashCode();	
	}
	
}