package org.netkernel.client.http.representation;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;

import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSBuilder;

public class HttpClientResponseRepresentation
{
	HttpResponse mResponse;
	IHDSNode mHDS;
	IReadableBinaryStreamRepresentation mEntity;
	
	public HttpClientResponseRepresentation(HttpResponse aResponse, Object aEntity)
	{	mResponse=aResponse;
		if(aEntity instanceof IReadableBinaryStreamRepresentation)
		{	mEntity=(IReadableBinaryStreamRepresentation)aEntity;
		}
	
		HDSBuilder builder = new HDSBuilder();
		builder.pushNode ("httpResult");
		builder.addNode ("resultCode", this.getResponseCode());
		builder.pushNode ("httpHeaders");
			Header[] headers=mResponse.getAllHeaders();
			for(Header h: headers)
			{	builder.addNode(h.getName(), h.getValue());
			}
		builder.popNode();
		builder.addNode ("entity", mEntity);
		mHDS=builder.getRoot();
	}
	
	public Map<String, String> getHeaders()
	{	Header[] hs=mResponse.getAllHeaders();
		HashMap<String, String> m=new HashMap();
		for(int i=0; i< hs.length; i++)
		{	Header h=hs[i];
			m.put(h.getName(), h.getValue());
		}
		return m;
	}
	
	public String getHeader(String header)
	{	Header h=mResponse.getFirstHeader(header);
		if(h==null)
		{	return null;
		}
		else return h.getValue();
	}

	public IHDSNode getImplementingHDS()
	{	return mHDS;
	}
	
	public IReadableBinaryStreamRepresentation getEntity()
	{	return mEntity;
	}
	
	public int getResponseCode()
	{	return mResponse.getStatusLine().getStatusCode();
	}
	
	protected HttpResponse getHttpResponseReadOnly()
	{	return mResponse;
	}

	@Override
	public boolean equals(Object obj)
	{	return (obj instanceof HttpClientResponseRepresentation) && ((HttpClientResponseRepresentation)obj).mHDS.equals(this.mHDS);
	}

	@Override
	public int hashCode()
	{	return mHDS.hashCode();
	}

	@Override
	public String toString()
	{	StringBuilder b=new StringBuilder();
		b.append("HttpClientResponseRepresentation\n");
		b.append(mHDS.toString());
		// create from readable binary stream
		String result=null;
		IReadableBinaryStreamRepresentation rbs = getEntity();
		int length = rbs.getContentLength();
		try
		{
			if (length>0)
			{	InputStream is=rbs.getInputStream();
				try
				{	
					InputStreamReader br = new InputStreamReader(is,rbs.getEncoding());
					char[] buffer = new char[length];
					int l=br.read(buffer,0, length);
					result = new String(buffer,0,l);
				}
				finally
				{	is.close();
				}
			}
		}
		catch(Exception e) { }
		b.append("\n");
		b.append(result);
		return b.toString();
	}
}
